==> nui_mini_program-1.0.3.tar.gz: lib/widget/nui_miniprogram_widget.dart <==
import 'dart:async';
import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_inappwebview/flutter_inappwebview.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_core/ui/component/toolbar/nui_status_bar.dart';
import 'package:nui_mini_program/engine/nui_miniprogram_js_engine.dart';
import 'package:nui_mini_program/model/nui_miniprogram_module.dart';
import 'package:nui_mini_program/util/nui_miniprogram_action_util.dart';
import 'package:nui_mini_program/util/nui_miniprogram_file_manager.dart';
import 'package:nui_miniprogram_core/data/nui_miniprogram_session.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_enum.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_message.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_engine.dart';
import 'package:nui_toolbox/common/nui_future_widget.dart';
import 'package:webview_flutter/webview_flutter.dart';
import 'package:webview_flutter_plus/webview_flutter_plus.dart';

typedef NUIMiniProgramSessionBuilder = Future<NUIMiniProgramSession> Function();
class NUIMiniProgramView extends StatefulWidget {
  NUIMiniProgramView({
    required this.module,
    required this.sessionBuilder,
    required this.controller,
    this.showStatusBar = true,
    this.placeHolder,
    this.loadUntilInitialized = false
  });

  final NUIMiniProgramModule module;
  final NUIMiniProgramController controller;
  final bool showStatusBar;
  final NUIMiniProgramSessionBuilder sessionBuilder;
  final WidgetBuilder? placeHolder;
  final bool loadUntilInitialized;

  @override
  _NUIMiniProgramViewState createState() => _NUIMiniProgramViewState();
}

class _NUIMiniProgramViewState extends State<NUIMiniProgramView> {
  static const MODULE = "NUIMiniProgramView";

  late NUIMiniProgramController _controller;
  late NUIMiniProgramFileManager _fileManager;
  late NUIMiniProgramJSEngine _jsEngine;
  late Future<String> _moduleIndexHtml;
  String? remoteUrl;
  late String _currentIndexFile;
  // final InAppLocalhostServer localhostServer = new InAppLocalhostServer();

  bool _firstLoaded = false;

  @override
  void initState(){
    super.initState();
    _controller = widget.controller;
    _fileManager = NUIMiniProgramFileManager(widget.module);
    _jsEngine = NUIMiniProgramJSEngine.from(widget.module);
    remoteUrl = widget.module.remoteUrl;
    _moduleIndexHtml = loadHtml();
    _controller._jsEngine = _jsEngine;
    _initChannels();
  }

  Future<String> loadHtml() async{
    // localhostServer.start();
    return _fileManager.initializeMiniProgram();
  }

  void _initChannels(){
    final channels = <NUIMiniProgramChannel>[];
    channels.addAll(_controller.channels);
    channels.add(NUIMiniProgramChannel(NUIMiniProgramMessage.DEFAULT_CHANNEL, listener: (message, msg) async{
      return NUIMiniProgramActionUtil().processAction(context, message, msg, initializeListener: (){
        if(widget.loadUntilInitialized == true && _firstLoaded != true) {
          setState(() {
            _firstLoaded = true;
          });
        }
      }, sessionBuilder: widget.sessionBuilder);
    }));
    _jsEngine.setChannels(channels);
  }

  @override
  void dispose(){
    super.dispose();
    if(!isNullOrEmpty(_currentIndexFile)) {
      _fileManager.disableDynamicAssets(_currentIndexFile);
    }
    _controller._jsEngine = null;
  }

  Future<bool> _overrideBackNavigation() async{
    final controller = await _controller.webController;
    final canGoBack = await controller.webViewController.canGoBack();
    if(canGoBack){
      controller.webViewController.goBack();
      return false;
    }
    else{
      final completer = Completer<bool>();

      NUIAsync.delayForResult(Duration(milliseconds: 300), () async => true).then((value) {
        if(!completer.isCompleted) {
          completer.complete(false);
        }
      });
      _controller.canMiniProgramPop().then((value) {
        if(!completer.isCompleted) {
          completer.complete(value);
        }
      });

      final canPop = await completer.future;
      final canMiniPop = canPop;
      if(canMiniPop){
        return false;
      }
      return true;
    }
  }

  /// Rendering the place holder
  Widget _loadingPlaceholder(BuildContext context){
    if(widget.placeHolder != null){
      return widget.placeHolder!(context);
    }
    if(!widget.module.getIconUrl().isEmpty) {
      return Center(child: loadImage(widget.module.getIconUrl(), width: 100, height: 100, placeHolder: SizedBox(width: 1, height: 1)));
    }
    else{
      return Center(child: CircularProgressIndicator());
    }
  }

  /// Render the mini-app through downloaded resource package
  Widget _renderBySource(){
    return NUIFutureWidget(
        futureValue: _moduleIndexHtml,
        callback: (state, html, message){
          if(state == FutureState.SUCCESS && !isNullOrEmpty(html)){
            html!; // Should not be null
            _currentIndexFile = html;
            final uri = Uri.directory(html);
            final uriString = uri.toString().substring(0, uri.toString().length - 1);
            _fileManager.enableDynamicAssets(html);
            logNUI(MODULE, "Rendering html from : $html, ${DateTime.now()}");
            return AnimatedOpacity(
                duration: Duration(milliseconds: 250),
                opacity: _firstLoaded ? 1 : 0,
                // child: getInAppWebView(html)
                child: getWebViewPlus(html)
            );
          }
          else{
            return SizedBox.shrink();
          }
        }
    );
  }

  /// Rendering the mini-app through remote url
  Widget _renderByRemote(){
    return getWebViewPlus(remoteUrl!);
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      resizeToAvoidBottomInset: false,
      body: WillPopScope(
        onWillPop: () async{
          return _overrideBackNavigation();
        },
        child: Container(
          width: double.infinity,
          height: double.infinity,
          decoration: NUIBoxDecoration(
            gradient: NUILinearGradient(
              colors: widget.module.getBrandColors() ?? [NUIColors.NUIWhite, NUIColors.NUISkeletalGray],
              orientation: GradientOrientation.TR_BL
            ),
          ),
          child: Column(
            mainAxisSize: MainAxisSize.max,
            children: [
              widget.showStatusBar == true ? NUIStatusBar() : SizedBox.shrink(),
              Expanded(
                child: Stack(
                  children: [
                    // AnimatedOpacity(
                    //   duration: Duration(milliseconds: 250),
                    //   opacity: _firstLoaded ? 0 : 1,
                    //   child: Container(
                    //     width: double.infinity,
                    //     height: double.infinity,
                    //     child: _loadingPlaceholder(context),
                    //   ),
                    // ),
                    remoteUrl != null ? _renderByRemote() : _renderBySource(),
                    AnimatedSwitcher(
                      duration: Duration(milliseconds: 250),
                      child: _firstLoaded ? SizedBox.shrink() : Container(
                        width: double.infinity,
                        height: double.infinity,
                        child: _loadingPlaceholder(context),
                      ),
                    ),
                  ],
                ),
              ),
            ],
          ),
        ),
      ),
    );
  }

  Widget getInAppWebView(String html){
    return InAppWebView(
      initialUrlRequest: URLRequest(url: Uri.parse("http://localhost:8080$html")),
    );
  }

  Widget getWebViewPlus(String html){
    return WebViewPlus(
        backgroundColor: (widget.module.getBrandColors()).first,
        initialUrl: html,
        javascriptMode: JavascriptMode.unrestricted,
        javascriptChannels: [
          _jsEngine.getJavascriptChannel("NUIMiniProgram"),
        ].toSet(),
        onWebViewCreated: (WebViewPlusController webViewController) {
          print("NUIMiniProgramView: Web view created");
          _jsEngine.setWebController(webViewController);
          _controller._setWebController(webViewController);
        },
        onPageStarted: (url){
          logNUI(MODULE, "On page started at url: $url");
        },
        onPageFinished: (url){
          if(widget.loadUntilInitialized != true && _firstLoaded != true) {
            setState(() {
              _firstLoaded = true;
            });
          }
          // logNUI(MODULE, "On page finished at url: $url");
        },
        gestureRecognizers: Set()
          ..add(
              Factory<LongPressGestureRecognizer>(() => LongPressGestureRecognizer()..onLongPress = (){
                logNUI("NUIMiniProgramView", "On long press detected");
              })
          )
    );
  }
}

/**
 * Mini-program controller
 */
class NUIMiniProgramController{
  static const MODULE = "NUIMiniProgramController";
  NUIMiniProgramController({List<NUIMiniProgramChannel>? channels}){
    _channels = channels ?? [];
  }

  NUIMiniProgramJSEngineInterface? _jsEngine;
  Completer<WebViewPlusController> _webController = Completer();
  late List<NUIMiniProgramChannel> _channels;

  Future<WebViewPlusController> get webController => _webController.future;

  Future<NUIMiniProgramMessage?> postMessage(NUIMiniProgramMessage message) async{
    try {
      await _webController.future;
      return _postMessageToMiniProgram(message);
    }catch(e){
      logNUI(MODULE, "Failed to encode message with error : $e");
      return null;
    }
  }
  
  Future<bool> canMiniProgramPop() async{
    final result = await _jsEngine?.sendAndAwaitCrossEvent(NUIMiniProgramMessage(channel: NUIMiniProgramWebEventType.CAN_POP.value, data: NUIMiniProgramWebEventType.CAN_POP.value));
    if(result != null){
      return match(result.data, "true");
    }
    return false;
  }

  Future<NUIMiniProgramMessage?> _postMessageToMiniProgram(NUIMiniProgramMessage message) async{
    final result = await _jsEngine?.sendAndAwaitCrossEvent(message);
    return result;
  }

  void _setWebController(WebViewPlusController controller){
    _webController.complete(controller);
  }

  List<NUIMiniProgramChannel> get channels => _channels ?? [];
}

==> nui_mini_program-1.0.5.tar.gz: lib/widget/nui_miniprogram_widget.dart <==
import 'dart:async';
import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_inappwebview/flutter_inappwebview.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_core/ui/component/toolbar/nui_status_bar.dart';
import 'package:nui_mini_program/engine/nui_miniprogram_js_engine.dart';
import 'package:nui_mini_program/model/nui_miniprogram_module.dart';
import 'package:nui_mini_program/util/nui_miniprogram_action_util.dart';
import 'package:nui_mini_program/util/nui_miniprogram_file_manager.dart';
import 'package:nui_miniprogram_core/data/nui_miniprogram_session.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_enum.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_message.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_engine.dart';
import 'package:nui_toolbox/common/nui_future_widget.dart';
import 'package:webview_flutter/webview_flutter.dart';
import 'package:webview_flutter_plus/webview_flutter_plus.dart';

typedef NUIMiniProgramSessionBuilder = Future<NUIMiniProgramSession> Function(); // Receiving the session request from mini-app and returning the session from the super-app
typedef NUIMiniProgramMessageExchangeBuilder = Future<dynamic> Function(dynamic message); // For custom message exchange between the super-app and the mini-app

/**
 * Mini-app container that will render and load the mini-app as an in-app web view
 */
class NUIMiniProgramView extends StatefulWidget {
  NUIMiniProgramView({
    required this.module,
    required this.controller,
    this.showStatusBar = true,
    this.placeHolder,
    this.loadUntilInitialized = false,

    required this.sessionBuilder,
    required this.messageExchange
  });

  final NUIMiniProgramModule module;
  final NUIMiniProgramController controller;
  final bool showStatusBar;
  final WidgetBuilder? placeHolder;
  final bool loadUntilInitialized;

  final NUIMiniProgramSessionBuilder sessionBuilder; // For session exchange
  final NUIMiniProgramMessageExchangeBuilder messageExchange; // For message exchange

  @override
  _NUIMiniProgramViewState createState() => _NUIMiniProgramViewState();
}

class _NUIMiniProgramViewState extends State<NUIMiniProgramView> {
  static const MODULE = "NUIMiniProgramView";

  late NUIMiniProgramController _controller;
  late NUIMiniProgramFileManager _fileManager;
  late NUIMiniProgramJSEngine _jsEngine;
  late Future<String> _moduleIndexHtml;
  String? remoteUrl;
  late String _currentIndexFile;
  // final InAppLocalhostServer localhostServer = new InAppLocalhostServer();

  bool _firstLoaded = false;

  @override
  void initState(){
    super.initState();
    _controller = widget.controller;
    _fileManager = NUIMiniProgramFileManager(widget.module);
    _jsEngine = NUIMiniProgramJSEngine.from(widget.module);
    remoteUrl = widget.module.remoteUrl;
    // _moduleIndexHtml = loadHtml();
    _controller._jsEngine = _jsEngine;
    _initChannels();
  }

  Future<String> loadHtml() async{
    // localhostServer.start();
    return _fileManager.initializeMiniProgram();
  }

  /// Initializing the channels
  void _initChannels(){
    final channels = <NUIMiniProgramChannel>[];

    // Adding the channels added from the controller
    channels.addAll(_controller.channels);

    // Adding the default channel
    channels.add(NUIMiniProgramChannel(NUIMiniProgramMessage.DEFAULT_CHANNEL, listener: (message, msg) async{
      return NUIMiniProgramActionUtil().processAction(context, message, msg, initializeListener: (){
        if(widget.loadUntilInitialized == true && _firstLoaded != true) {
          setState(() {
            _firstLoaded = true;
          });
        }
      }, sessionBuilder: widget.sessionBuilder, messageExchangeBuilder: widget.messageExchange);
    }));
    _jsEngine.setChannels(channels);
  }

  @override
  void dispose(){
    super.dispose();
    if(!isNullOrEmpty(_currentIndexFile)) {
      _fileManager.disableDynamicAssets(_currentIndexFile);
    }
    _controller._jsEngine = null;
  }

  /// Check if hte back navigation should be overriden by the mini-app navigation
  Future<bool> _overrideBackNavigation() async{
    final controller = await _controller.webController;
    final canGoBack = await controller.webViewController.canGoBack();
    if(canGoBack){
      controller.webViewController.goBack();
      return false;
    }
    else{
      final completer = Completer<bool>();

      NUIAsync.delayForResult(Duration(milliseconds: 300), () async => true).then((value) {
        if(!completer.isCompleted) {
          completer.complete(false);
        }
      });
      _controller.canMiniProgramPop().then((value) {
        if(!completer.isCompleted) {
          completer.complete(value);
        }
      });

      final canPop = await completer.future;
      final canMiniPop = canPop;
      if(canMiniPop){
        return false;
      }
      return true;
    }
  }

  /// Rendering the place holder
  Widget _loadingPlaceholder(BuildContext context){
    if(widget.placeHolder != null){
      return widget.placeHolder!(context);
    }
    if(!widget.module.getIconUrl().isEmpty) {
      return Center(child: loadImage(widget.module.getIconUrl(), width: 100, height: 100, placeHolder: SizedBox(width: 1, height: 1)));
    }
    else{
      return Center(child: CircularProgressIndicator());
    }
  }

  /// Render the mini-app through downloaded resource package
  Widget _renderBySource(){
    return NUIFutureWidget(
        futureValue: _moduleIndexHtml,
        callback: (state, html, message){
          if(state == FutureState.SUCCESS && !isNullOrEmpty(html)){
            html!; // Should not be null
            _currentIndexFile = html;
            final uri = Uri.directory(html);
            final uriString = uri.toString().substring(0, uri.toString().length - 1);
            _fileManager.enableDynamicAssets(html);
            logNUI(MODULE, "Rendering html from : $html, ${DateTime.now()}");
            return AnimatedOpacity(
                duration: Duration(milliseconds: 250),
                opacity: _firstLoaded ? 1 : 0,
                // child: getInAppWebView(html)
                child: getWebViewPlus(html)
            );
          }
          else{
            return SizedBox.shrink();
          }
        }
    );
  }

  /// Rendering the mini-app through remote url
  Widget _renderByRemote(){
    return getWebViewPlus(remoteUrl!);
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      resizeToAvoidBottomInset: false,
      body: WillPopScope(
        onWillPop: () async{
          return _overrideBackNavigation();
        },
        child: Container(
          width: double.infinity,
          height: double.infinity,
          decoration: NUIBoxDecoration(
            gradient: NUILinearGradient(
              colors: widget.module.getBrandColors() ?? [NUIColors.NUIWhite, NUIColors.NUISkeletalGray],
              orientation: GradientOrientation.TR_BL
            ),
          ),
          child: Column(
            mainAxisSize: MainAxisSize.max,
            children: [
              widget.showStatusBar == true ? NUIStatusBar() : SizedBox.shrink(), // Show/hide the status bar
              Expanded(
                child: Stack(
                  children: [
                    remoteUrl != null ? _renderByRemote() : _renderBySource(),
                    AnimatedSwitcher(
                      duration: Duration(milliseconds: 250),
                      child: _firstLoaded ? SizedBox.shrink() : Container(
                        width: double.infinity,
                        height: double.infinity,
                        child: _loadingPlaceholder(context),
                      ),
                    ),
                  ],
                ),
              ),
            ],
          ),
        ),
      ),
    );
  }

  /// Getting the in-app web view
  Widget getInAppWebView(String html){
    return InAppWebView(
      initialUrlRequest: URLRequest(url: Uri.parse("http://localhost:8080$html")),
    );
  }

  /// Retrieving the webview plus instance
  Widget getWebViewPlus(String html){
    return WebViewPlus(
        backgroundColor: (widget.module.getBrandColors()).first,
        initialUrl: html,
        javascriptMode: JavascriptMode.unrestricted,
        javascriptChannels: [
          _jsEngine.getJavascriptChannel("NUIMiniProgram"),
        ].toSet(),
        onWebViewCreated: (WebViewPlusController webViewController) {
          print("NUIMiniProgramView: Web view created");
          _jsEngine.setWebController(webViewController);
          _controller._setWebController(webViewController);
        },
        onPageStarted: (url){
          logNUI(MODULE, "On page started at url: $url");
        },
        onPageFinished: (url){
          if(widget.loadUntilInitialized != true && _firstLoaded != true) {
            setState(() {
              _firstLoaded = true;
            });
          }
          // logNUI(MODULE, "On page finished at url: $url");
        },
        gestureRecognizers: Set()
          ..add(
              Factory<LongPressGestureRecognizer>(() => LongPressGestureRecognizer()..onLongPress = (){
                logNUI("NUIMiniProgramView", "On long press detected");
              })
          )
    );
  }
}

/**
 * Mini-program controller
 */
class NUIMiniProgramController{
  static const MODULE = "NUIMiniProgramController";
  NUIMiniProgramController({List<NUIMiniProgramChannel>? channels}){
    _channels = channels ?? [];
  }

  NUIMiniProgramJSEngineInterface? _jsEngine;
  Completer<WebViewPlusController> _webController = Completer();
  late List<NUIMiniProgramChannel> _channels;

  Future<WebViewPlusController> get webController => _webController.future;

  /// Posting a message to the mini-app
  Future<NUIMiniProgramMessage?> postMessage(NUIMiniProgramMessage message) async{
    try {
      await _webController.future;
      return _postMessageToMiniProgram(message);
    }catch(e){
      logNUI(MODULE, "Failed to encode message with error : $e");
      return null;
    }
  }

  /// Sending custom message
  Future<dynamic> exchangeMessage(dynamic input) async{
    final message = NUIMiniProgramMessage(channel: NUIMiniProgramMessage.DEFAULT_CHANNEL, data: input);
    final result = await _jsEngine?.sendAndAwaitCrossEvent(message);
    return result?.data;
  }

  /// Checking if the mini-app can pop/close
  Future<bool> canMiniProgramPop() async{
    final result = await _jsEngine?.sendAndAwaitCrossEvent(NUIMiniProgramMessage(channel: NUIMiniProgramWebEventType.CAN_POP.value, data: NUIMiniProgramWebEventType.CAN_POP.value));
    if(result != null){
      return match(result.data, "true");
    }
    return false;
  }

  /// Sending message to the mini-app
  Future<NUIMiniProgramMessage?> _postMessageToMiniProgram(NUIMiniProgramMessage message) async{
    final result = await _jsEngine?.sendAndAwaitCrossEvent(message);
    return result;
  }

  void _setWebController(WebViewPlusController controller){
    _webController.complete(controller);
  }

  List<NUIMiniProgramChannel> get channels => _channels ?? [];
}
